/* ************************************************************************** */
/*                                       _____           _                    */
/*  core/init_usb.c                     |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/09/05 17:02:00                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <libp7/internals.h>
#include <libusb-1.0/libusb.h>
#include <unistd.h>

/* USB identifiers */
#define casio_vendor 0x07CF
#define casio_product 0x6101

/* libusb endpoints */
#define ENDPOINT_OUT (LIBUSB_ENDPOINT_OUT | LIBUSB_TRANSFER_TYPE_ISOCHRONOUS)
#define ENDPOINT_IN (LIBUSB_ENDPOINT_IN | LIBUSB_TRANSFER_TYPE_BULK)

/* ************************************************************************** */
/*  Stream utilities                                                          */
/* ************************************************************************** */
/**
 *	cookie_t:
 *	The cookie type.
 */

typedef struct {
	libusb_context *context;
	libusb_device_handle *handle;
} cookie_t;

/**
 *	usb_read:
 *	Read packet.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	sbuf		the buffer pointer (signed)
 *	@arg	size		the size to read
 *	@return				read bytes count or error
 */

static ssize_t usb_read(void *vcookie, char *sbuf, size_t size)
{
	/* cast cookie and buffer */
	cookie_t *cookie = vcookie;
	unsigned char *buf = (unsigned char*)sbuf;

	/* last read bytes, deduce timeout of it */
	long timeout = (p7_active) ? ACTIVE_TIMEOUT : PASSVE_TIMEOUT;
	timeout *= 1000; /* milliseconds -> microseconds */

	/* get the end */
	long end = p7_get_ms() + timeout;

	/* read bytes */
	int received = 0, err;
	while (!received) {
		/* check time left */
		long left = end - p7_get_ms();
		if (left <= 0) {
			errno = ETIME;
			return (-1);
		}

		/* receive */
		unsigned int leftms = left / 1000;
		err = libusb_bulk_transfer(cookie->handle, ENDPOINT_IN, buf,
			size, &received, leftms);

		/* check error */
		switch (err) {
		/* - timeout - */
		case LIBUSB_ERROR_TIMEOUT:
			errno = ETIME;
			return (-1);

		/* - device error - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			errno = ENODEV;
			return (-1);

		/* - overflow - */
		case LIBUSB_ERROR_OVERFLOW:
			/* no error - we will simply read what's next */
			break;

		case LIBUSB_ERROR_IO:
			errno = EREMOTEIO;
			return (-1);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb error was %d : %s", err, libusb_strerror(err));
			errno = EPERM;
			return (-1);
		}
	}

	/* return transferred size */
	return ((ssize_t)received);
}

/**
 *	usb_write:
 *	Send packet.
 *
 *	@arg	vcookie		the cookie (uncasted)
 *	@arg	sbuf		the buffer (signed)
 *	@arg	size		the size of the buffer
 *	@return				number of bytes written
 */

static ssize_t usb_write(void *vcookie, const char *sbuf, size_t size)
{
	/* cast cookie and buffer */
	cookie_t *cookie = vcookie;
	unsigned char *buf = (const unsigned char*)sbuf;

	/* write */
	int tf = 0, err;
	err = libusb_bulk_transfer(cookie->handle, ENDPOINT_OUT, buf,
		size, &tf, SENDNG_TIMEOUT);

	/* check error */
	switch (err) {
	/* - timeout error - */
	case LIBUSB_ERROR_TIMEOUT:
		errno = ETIME;
		return (-1);

	/* - device error - */
	case LIBUSB_ERROR_PIPE:
	case LIBUSB_ERROR_NO_DEVICE:
	case LIBUSB_ERROR_OVERFLOW:
		errno = ENODEV;
		return (-1);

	/* - unknown error - */
	case 0: break;
	default:
		errno = EPERM;
		return (-1);
	}

	/* return transferred size */
	return ((ssize_t)tf);
}

/**
 *	usb_close:
 *	Close function.
 *
 *	@arg	vcookie		cookie (uncasted)
 *	@return				if everything went fine
 */

static int usb_close(void *vcookie)
{
	/* cast cookie */
	cookie_t *cookie = vcookie;

	/* free cookie!!! order now! */
	if (cookie->handle) libusb_close(cookie->handle);
	if (cookie->context) libusb_exit(cookie->context);

	/* everything went well */
	return (0);
}

/* ************************************************************************** */
/*  Main function                                                             */
/* ************************************************************************** */
/**
 *	p7_init:
 *	Initialize libp7 with USB.
 *
 *	@arg	active	start as active ?
 *	@return			if initialization went okay.
 */

int p7_init(int active)
{
	/* bake vars */
	static cookie_t cookie = {NULL, NULL};
	FILE *stream = NULL;

	/* open handle */
	log_info("creating libusb context");
	if (libusb_init(&cookie.context)) {
		log_fatal("couldn't create libusb context");
		p7_error = p7_error_libusb;
		goto fail;
	}

	/* checking for the calculator */
	int tries = 10;
	libusb_device **device_list = NULL;
	libusb_device *calc = NULL;
	log_info("checking for the calculator on USB connexions");
	while (tries && !calc) {
		/* - get device list - */
		int device_count = libusb_get_device_list(cookie.context, &device_list);
		if (device_count < 0) {
			log_error("could not get device list :(");
			p7_error = p7_error_unknown;
			goto fail;
		}
		/* - looking for the calc - */
		for (int id = 0; id < device_count; id++) {
			struct libusb_device_descriptor descriptor;
			if (libusb_get_device_descriptor(device_list[id], &descriptor))
				continue;
			if (descriptor.idVendor == casio_vendor
			 && descriptor.idProduct == casio_product) {
				calc = device_list[id];
				break;
			}
		}
		/* - if no calc, sleep and try again - */
		if (!calc) {
			log_info("could not find calculator, sleep and try again");
			tries--; /* do not put that in log! */
			log_info("we have %d tries left.", tries);

			/* free list and sleep before trying again */
			libusb_free_device_list(device_list, 1);
			device_list = NULL;
			if (tries) usleep(400000);
		}
	}
	/* - if no calc, snif - */
	if (!calc) {
		log_error("looks like we couldn't find the calculator");
		libusb_free_device_list(device_list, 1);
		p7_error = p7_error_nocalc;
		goto fail;
	}

	/* get calculator handler */
	log_info("getting the device handler");
	int err = libusb_open(calc, &cookie.handle);
	libusb_free_device_list(device_list, 1);
	switch (err) {
	/* couldn't get access to device */
	case LIBUSB_ERROR_ACCESS:
		p7_error = p7_error_noaccess;
		goto fail;

	/* default case */
	case 0: break;
	default:
		log_fatal("libusb_open returned %d: %s", err,
			libusb_error_name(err));
		log_fatal("errno was %d: %s", errno, strerror(errno));
		p7_error = p7_error_unknown;
		goto fail;
	}

	/* create filestream */
	stream = fopencookie(&cookie, "r+",
		(cookie_io_functions_t){&usb_read, &usb_write, NULL, &usb_close});
	if (!stream)
		goto fail;

	/* initialize raw */
	return (p7_init_stream(stream, active));
fail:
	usb_close((void*)&cookie);
	return (0);
}
